import numpy as np
from numpy.typing import ArrayLike
from scipy import spatial
from scipy.spatial import KDTree
from scipy.sparse.csgraph import connected_components

from .coordinates import rotate_axis, polar_coordinates, Coordinates, CoordinateFrame
from .atoms import next_neighbors
from .autosave import autosave_data
from .utils import runningmean
from .pbc import pbc_diff, pbc_points
from .logging import logger


@autosave_data(nargs=2, kwargs_keys=("coordinates_b",), version="time_average-1")
def time_average(function, coordinates, coordinates_b=None, pool=None):
    """
    Compute the time average of a function.

    Args:
        function:
            The function that will be averaged, it has to accept exactly one argument
            which is the current atom set
        coordinates: The coordinates object of the simulation
        pool (multiprocessing.Pool, opt.):
            A multiprocessing pool which will be used for cocurrent calculation of the
            averaged function

    """
    if pool is not None:
        _map = pool.imap
    else:
        _map = map

    number_of_averages = 0
    result = 0

    if coordinates_b is not None:
        if coordinates._slice != coordinates_b._slice:
            logger.warning("Different slice for coordinates and coordinates_b.")
        coordinate_iter = (iter(coordinates), iter(coordinates_b))
    else:
        coordinate_iter = (iter(coordinates),)

    evaluated = _map(function, *coordinate_iter)

    for ev in evaluated:
        number_of_averages += 1
        result += ev
        if number_of_averages % 100 == 0:
            logger.debug("time_average: %d", number_of_averages)

    return result / number_of_averages


def time_histogram(function, coordinates, bins, hist_range, pool=None):
    coordinate_iter = iter(coordinates)

    if pool is not None:
        _map = pool.imap
    else:
        _map = map

    evaluated = _map(function, coordinate_iter)

    results = []
    hist_results = []
    for num, ev in enumerate(evaluated):
        results.append(ev)

        if num % 100 == 0 and num > 0:
            print(num)
        r = np.array(results).T
        for i, row in enumerate(r):
            histo, _ = np.histogram(row, bins=bins, range=hist_range)
            if len(hist_results) <= i:
                hist_results.append(histo)
            else:
                hist_results[i] += histo
            results = []
    return hist_results


def calc_gr(
    traj_a: Coordinates,
    traj_b: Coordinates = None,
    r_max: float = None,
    delta_r: float = 0.02,
    segments: int = 1000,
    skip: float = 0.1,
    remove_intra: bool = False,
    shear: bool = False,
):
    r"""
    Compute the radial pair distribution of one or two sets of atoms.

    .. math::
       g_{AB}(r) = \frac{1}{\langle \rho_B\rangle N_A}\sum\limits_{i\in A}^{N_A}
       \sum\limits_{j\in B}^{N_B}\frac{\delta(r_{ij} -r)}{4\pi r^2}

    For use with :func:`time_average`, define bins through the use of :func:`~functools.partial`,
    the atom sets are passed to :func:`time_average`, if a second set of atoms should be used
    specify it as ``coordinates_b`` and it will be passed to this function.

    Args:
        atoms_a: First set of atoms, used internally
        atoms_b (opt.): Second set of atoms, used internally
        bins: Bins of the radial distribution function
        box (opt.): Simulations box, if not specified this is taken from ``atoms_a.box``
        kind (opt.): Can be 'inter', 'intra' or None (default).
        chunksize (opt.):
            For large systems (N > 1000) the distaces have to be computed in chunks so the arrays
            fit into memory, this parameter controlls the size of these chunks. It should be
            as large as possible, depending on the available memory.
        returnx (opt.): If True the x ordinate of the histogram is returned.
    """

    def gr_frame(
        atoms_a: CoordinateFrame,
        atoms_b: CoordinateFrame,
        bins: ArrayLike,
        remove_intra: bool = False,
    ):
        box = atoms_b.box
        n = len(atoms_a) / np.prod(np.diag(box))
        V = 4 / 3 * np.pi * bins[-1] ** 3
        particles_in_volume = int(n * V * 1.1)
        if np.all(np.diag(np.diag(box)) == box):
            atoms_b = atoms_b % np.diag(box)
            atoms_b_res_ids = atoms_b.residue_ids
            atoms_b_tree = KDTree(atoms_b, boxsize=np.diag(box))
        else:
            atoms_b_pbc, atoms_b_pbc_index = pbc_points(
                atoms_b, box, thickness=bins[-1] + 0.1, index=True, shear=shear
            )
            atoms_b_res_ids = atoms_b.residue_ids[atoms_b_pbc_index]
            atoms_b_tree = KDTree(atoms_b_pbc)
        distances, distances_index = atoms_b_tree.query(
            atoms_a, particles_in_volume, distance_upper_bound=bins[-1] + 0.1
        )
        if np.array_equal(atoms_a, atoms_b):
            distances = distances[:, 1:]
            distances_index = distances_index[:, 1:]

        if remove_intra:
            new_distances = []
            for entry in list(zip(atoms_a.residue_ids, distances, distances_index)):
                mask = entry[1] < np.inf
                new_distances.append(
                    entry[1][mask][atoms_b_res_ids[entry[2][mask]] != entry[0]]
                )
            distances = np.concatenate(new_distances)
        else:
            distances = distances.flatten()

        hist = np.histogram(distances, bins=bins, range=(0, bins[-1]), density=False)[0]
        gr = hist / len(atoms_a)
        gr = gr / (4 / 3 * np.pi * bins[1:] ** 3 - 4 / 3 * np.pi * bins[:-1] ** 3)
        n = len(atoms_b) / np.prod(np.diag(atoms_b.box))
        gr = gr / n

        return gr, n

    if traj_b is None:
        traj_b = traj_a

    start_frame = traj_a[int(len(traj_a) * skip)]
    if r_max:
        upper_bound = r_max
    else:
        upper_bound = round(np.min(np.diag(start_frame.box)) / 2 - 0.05, 1)

    num_steps = int(upper_bound * (1 / delta_r) + 1)
    bins = np.linspace(0, upper_bound, num_steps)
    r = bins[1:] - (bins[1] - bins[0]) / 2
    frame_indices = np.unique(
        np.int_(np.linspace(len(traj_a) * skip, len(traj_a) - 1, num=segments))
    )
    gr = []
    n = []
    for frame_index in frame_indices:
        result = gr_frame(
            traj_a[frame_index], traj_b[frame_index], bins, remove_intra=remove_intra
        )
        gr.append(result[0])
        n.append(result[1])
    gr = np.mean(gr, axis=0)
    n = np.mean(n, axis=0)
    return r, gr, n


def distance_distribution(atoms, bins):
    connection_vectors = atoms[:-1, :] - atoms[1:, :]
    connection_lengths = (connection_vectors**2).sum(axis=1) ** 0.5
    return np.histogram(connection_lengths, bins)[0]


def tetrahedral_order(atoms, reference_atoms=None):
    if reference_atoms is None:
        reference_atoms = atoms
    indices = next_neighbors(reference_atoms, query_atoms=atoms, number_of_neighbors=4)
    neighbors = reference_atoms[indices]
    neighbors_1, neighbors_2, neighbors_3, neighbors_4 = (
        neighbors[:, 0, :],
        neighbors[:, 1, :],
        neighbors[:, 2, :],
        neighbors[:, 3, :],
    )

    # Connection vectors
    neighbors_1 -= atoms
    neighbors_2 -= atoms
    neighbors_3 -= atoms
    neighbors_4 -= atoms

    # Normed Connection vectors
    neighbors_1 /= np.linalg.norm(neighbors_1, axis=-1).reshape(-1, 1)
    neighbors_2 /= np.linalg.norm(neighbors_2, axis=-1).reshape(-1, 1)
    neighbors_3 /= np.linalg.norm(neighbors_3, axis=-1).reshape(-1, 1)
    neighbors_4 /= np.linalg.norm(neighbors_4, axis=-1).reshape(-1, 1)

    a_1_2 = ((neighbors_1 * neighbors_2).sum(axis=1) + 1 / 3) ** 2
    a_1_3 = ((neighbors_1 * neighbors_3).sum(axis=1) + 1 / 3) ** 2
    a_1_4 = ((neighbors_1 * neighbors_4).sum(axis=1) + 1 / 3) ** 2

    a_2_3 = ((neighbors_2 * neighbors_3).sum(axis=1) + 1 / 3) ** 2
    a_2_4 = ((neighbors_2 * neighbors_4).sum(axis=1) + 1 / 3) ** 2

    a_3_4 = ((neighbors_3 * neighbors_4).sum(axis=1) + 1 / 3) ** 2

    q = 1 - 3 / 8 * (a_1_2 + a_1_3 + a_1_4 + a_2_3 + a_2_4 + a_3_4)

    return q


def tetrahedral_order_distribution(atoms, reference_atoms=None, bins=None):
    assert bins is not None, "Bin edges of the distribution have to be specified."
    Q = tetrahedral_order(atoms, reference_atoms=reference_atoms)
    return np.histogram(Q, bins=bins)[0]


def radial_density(
    atoms, bins, symmetry_axis=(0, 0, 1), origin=(0, 0, 0), height=1, returnx=False
):
    """
    Calculate the radial density distribution.

    This function is meant to be used with time_average.

    Args:
        atoms:
            Set of coordinates.
        bins:
            Bin specification that is passed to numpy.histogram. This needs to be
            a list of bin edges if the function is used within time_average.
        symmetry_axis (opt.):
            Vector of the symmetry axis, around which the radial density is calculated,
            default is z-axis.
        origin (opt.):
            Origin of the rotational symmetry, e.g. center of the pore.
        height (opt.):
            Height of the pore, necessary for correct normalization of the density.
        returnx (opt.):
            If True, the x ordinate of the distribution is returned.
    """
    cartesian = rotate_axis(atoms - origin, symmetry_axis)
    radius, _ = polar_coordinates(cartesian[:, 0], cartesian[:, 1])
    hist = np.histogram(radius, bins=bins)[0]
    volume = np.pi * (bins[1:] ** 2 - bins[:-1] ** 2) * height
    res = hist / volume
    if returnx:
        return np.vstack((runningmean(bins, 2), res))
    else:
        return res


def shell_density(
    atoms,
    shell_radius,
    bins,
    shell_thickness=0.5,
    symmetry_axis=(0, 0, 1),
    origin=(0, 0, 0),
):
    """
    Compute the density distribution on a cylindrical shell.

    Args:
        atoms: The coordinates of the atoms
        shell_radius: Inner radius of the shell
        bins: Histogramm bins, this has to be a two-dimensional list of bins: [angle, z]
        shell_thickness (opt.): Thicknes of the shell, default is 0.5
        symmetry_axis (opt.): The symmtery axis of the pore, the coordinates will be
            rotated such that this axis is the z-axis
        origin (opt.): Origin of the pore, the coordinates will be moved such that this
            is the new origin.

    Returns:
        Two-dimensional density distribution of the atoms in the defined shell.
    """
    cartesian = rotate_axis(atoms - origin, symmetry_axis)
    radius, theta = polar_coordinates(cartesian[:, 0], cartesian[:, 1])
    shell_indices = (shell_radius <= radius) & (
        radius <= shell_radius + shell_thickness
    )
    hist = np.histogram2d(theta[shell_indices], cartesian[shell_indices, 2], bins)[0]

    return hist


def spatial_density(atoms, bins, weights=None):
    """
    Compute the spatial density distribution.
    """
    density, _ = np.histogramdd(atoms, bins=bins, weights=weights)
    return density


def mixing_ratio_distribution(
    atoms_a,
    atoms_b,
    bins_ratio,
    bins_density,
    weights_a=None,
    weights_b=None,
    weights_ratio=None,
):
    """
    Compute the distribution of the mixing ratio of two sets of atoms.
    """

    density_a, _ = time_average
    density_b, _ = np.histogramdd(atoms_b, bins=bins_density, weights=weights_b)
    mixing_ratio = density_a / (density_a + density_b)
    good_inds = (density_a != 0) & (density_b != 0)
    hist, _ = np.histogram(
        mixing_ratio[good_inds], bins=bins_ratio, weights=weights_ratio
    )
    return hist


def next_neighbor_distribution(
    atoms, reference=None, number_of_neighbors=4, bins=None, normed=True
):
    """
    Compute the distribution of next neighbors with the same residue name.
    """
    assert bins is not None, "Bins have to be specified."
    if reference is None:
        reference = atoms
    nn = next_neighbors(
        reference, query_atoms=atoms, number_of_neighbors=number_of_neighbors
    )
    resname_nn = reference.residue_names[nn]
    count_nn = (resname_nn == atoms.residue_names.reshape(-1, 1)).sum(axis=1)
    return np.histogram(count_nn, bins=bins, normed=normed)[0]


def hbonds(
    D,
    H,
    A,
    box,
    DA_lim=0.35,
    HA_lim=0.35,
    min_cos=np.cos(30 * np.pi / 180),
    full_output=False,
):
    """
    Compute h-bond pairs

    Args:
        D: Set of coordinates for donators.
        H: Set of coordinates for hydrogen atoms. Should have the same
            length as D.
        A:  Set of coordinates for acceptors.
        DA_lim (opt.): Minimum distance beteen donator and acceptor.
        HA_lim (opt.): Minimum distance beteen hydrogen and acceptor.
        min_cos (opt.): Minimum cosine for the HDA angle. Default is
            equivalent to a maximum angle of 30 degree.
        full_output (opt.): Returns additionally the cosine of the
        angles and the DA distances

    Return:
        List of (D,A)-pairs in hbonds.
    """

    def dist_DltA(D, H, A, box, max_dist=0.35):
        ppoints, pind = pbc_points(D, box, thickness=max_dist + 0.1, index=True)
        Dtree = spatial.cKDTree(ppoints)
        Atree = spatial.cKDTree(A)
        pairs = Dtree.sparse_distance_matrix(Atree, max_dist, output_type="ndarray")
        pairs = np.asarray(pairs.tolist())
        pairs = np.int_(pairs[pairs[:, 2] > 0][:, :2])
        pairs[:, 0] = pind[pairs[:, 0]]
        return pairs

    def dist_AltD(D, H, A, box, max_dist=0.35):
        ppoints, pind = pbc_points(A, box, thickness=max_dist + 0.1, index=True)
        Atree = spatial.cKDTree(ppoints)
        Dtree = spatial.cKDTree(D)
        pairs = Atree.sparse_distance_matrix(Dtree, max_dist, output_type="ndarray")
        pairs = np.asarray(pairs.tolist())
        pairs = np.int_(pairs[pairs[:, 2] > 0][:, :2])
        pairs = pairs[:, ::-1]
        pairs[:, 1] = pind[pairs[:, 1]]
        return pairs

    if len(D) <= len(A):
        pairs = dist_DltA(D, H, A, box, DA_lim)
    else:
        pairs = dist_AltD(D, H, A, box, DA_lim)

    vDH = pbc_diff(D[pairs[:, 0]], H[pairs[:, 0]], box)
    vDA = pbc_diff(D[pairs[:, 0]], A[pairs[:, 1]], box)
    vHA = pbc_diff(H[pairs[:, 0]], A[pairs[:, 1]], box)
    angles_cos = np.clip(
        np.einsum("ij,ij->i", vDH, vDA)
        / np.linalg.norm(vDH, axis=1)
        / np.linalg.norm(vDA, axis=1),
        -1,
        1,
    )
    is_bond = (
        (angles_cos >= min_cos)
        & (np.sum(vHA**2, axis=-1) <= HA_lim**2)
        & (np.sum(vDA**2, axis=-1) <= DA_lim**2)
    )
    if full_output:
        return (
            pairs[is_bond],
            angles_cos[is_bond],
            np.sum(vDA[is_bond] ** 2, axis=-1) ** 0.5,
        )
    else:
        return pairs[is_bond]


def calc_cluster_sizes(frame, r_max=0.35):
    frame_PBC, indices_PBC = pbc_points(
        frame, frame.box, thickness=r_max + 0.1, index=True
    )
    tree = KDTree(frame_PBC)
    matrix = tree.sparse_distance_matrix(tree, r_max, output_type="ndarray")
    new_matrix = np.zeros((len(frame), len(frame)))
    for entry in matrix:
        if entry[2] > 0:
            new_matrix[indices_PBC[entry[0]], indices_PBC[entry[1]]] = 1
    n_components, labels = connected_components(new_matrix, directed=False)
    cluster_sizes = []
    for i in range(0, np.max(labels) + 1):
        cluster_sizes.append(np.sum(labels == i))
    return np.array(cluster_sizes).flatten()
