import numpy as np
from scipy.spatial import KDTree
from scipy.special import sph_harm


def a_ij(atoms, N=4, l=3):
    tree = KDTree(atoms)

    dist, indices = tree.query(atoms, N + 1)
    indices = indices[:, 1:]

    vecs = atoms[:, np.newaxis, :] - atoms[indices]
    vecs /= np.linalg.norm(vecs, axis=-1)[..., np.newaxis]

    theta = np.arctan2(vecs[..., 1], vecs[..., 0]) + np.pi
    phi = np.arccos(vecs[..., 2])
    qijlm = sph_harm(
        np.arange(-l, l + 1)[np.newaxis, np.newaxis, :],
        l,
        theta[..., np.newaxis],
        phi[..., np.newaxis],
    )
    qilm = np.average(qijlm, axis=1)
    qil = np.sum(qilm * np.conj(qilm), axis=-1) ** 0.5
    aij = (
        np.sum(qilm[:, np.newaxis, :] * np.conj(qilm[indices]), axis=-1)
        / qil[:, np.newaxis]
        / qil[indices]
    )
    return np.real(aij), indices


def number_of_neighbors(atoms):
    tree = KDTree(atoms)
    dist, _ = tree.query(atoms, 10, distance_upper_bound=0.35)
    return np.array([len(distance[distance < 0.4]) - 1 for distance in dist])


def classify_ice(aij, indices, neighbors, indexSOL):
    staggerdBonds = np.sum(aij <= -0.8, axis=1)
    eclipsedBonds = np.sum((aij >= -0.35) & (aij <= 0.25), axis=1)

    iceTypes = np.full(len(aij), 5)
    for i in indexSOL:
        if neighbors[i] != 4:
            continue
        elif staggerdBonds[i] == 4:
            iceTypes[i] = 0
        elif staggerdBonds[i] == 3 and eclipsedBonds[i] == 1:
            iceTypes[i] = 1
        elif staggerdBonds[i] == 3:
            for j in indices[i]:
                if staggerdBonds[j] >= 2:
                    iceTypes[i] = 2
                    break
        elif staggerdBonds[i] == 2:
            for j in indices[i]:
                if staggerdBonds[j] >= 3:
                    iceTypes[i] = 2
                    break
        elif eclipsedBonds[i] == 4:
            iceTypes[i] = 3
        elif eclipsedBonds[i] == 3:
            iceTypes[i] = 4
    iceTypes = iceTypes[indexSOL]

    return iceTypes


def ice_parts(iceTypes):
    cubic = len(iceTypes[iceTypes == 0])
    hexagonal = len(iceTypes[iceTypes == 1])
    interface = len(iceTypes[iceTypes == 2])
    clathrate = len(iceTypes[iceTypes == 3])
    clathrate_interface = len(iceTypes[iceTypes == 4])
    liquid = len(iceTypes[iceTypes == 5])
    return np.array(
        [cubic, hexagonal, interface, clathrate, clathrate_interface, liquid]
    )
