import numpy as np
from scipy.special import gamma as spgamma
from scipy.integrate import quad as spquad


def kww(t, A, tau, beta):
    return A * np.exp(-((t / tau) ** beta))


def kww_1e(A, tau, beta):
    return tau * (-np.log(1 / (np.e * A))) ** (1 / beta)


def cole_davidson(omega, A, beta, tau):
    P = np.arctan(omega * tau)
    return A * np.cos(P) ** beta * np.sin(beta * P)


def cole_cole(omega, A, beta, tau):
    return (
        A
        * (omega * tau) ** beta
        * np.sin(np.pi * beta / 2)
        / (
            1
            + 2 * (omega * tau) ** beta * np.cos(np.pi * beta / 2)
            + (omega * tau) ** (2 * beta)
        )
    )


def havriliak_negami(omega, A, beta, alpha, tau):
    r"""
    Imaginary part of the Havriliak-Negami function.

    .. math::
       \chi_{HN}(\omega) = \Im\left(\frac{A}{(1 + (i\omega\tau)^\alpha)^\beta}\right)
    """
    return -(A / (1 + (1j * omega * tau) ** alpha) ** beta).imag


# fits decay of correlation times, e.g. with distance to pore walls
def colen(d, X, t8, A):
    return t8 * np.exp(A * np.exp(-d / X))


# fits decay of the plateau height of the overlap function,
# e.g. with distance to pore walls
def colenQ(d, X, Qb, g):
    return (1 - Qb) * np.exp(-((d / X) ** g)) + Qb


def vft(T, tau_0, B, T_inf):
    return tau_0 * np.exp(B / (T - T_inf))


def arrhenius(T, tau_0, E_a):
    return tau_0 * np.exp(E_a / T)


def MLfit(t, tau, A, alpha):
    def MLf(z, a):
        """Mittag-Leffler function"""
        z = np.atleast_1d(z)
        if a == 0:
            return 1 / (1 - z)
        elif a == 1:
            return np.exp(z)
        elif a > 1 or all(z > 0):
            k = np.arange(100)
            return np.polynomial.polynomial.polyval(z, 1 / spgamma(a * k + 1))

        # a helper for tricky case, from Gorenflo, Loutchko & Luchko
        def _MLf(z, a):
            if z < 0:
                f = lambda x: (
                    np.exp(-x * (-z) ** (1 / a))
                    * x ** (a - 1)
                    * np.sin(np.pi * a)
                    / (x ** (2 * a) + 2 * x**a * np.cos(np.pi * a) + 1)
                )
                return 1 / np.pi * spquad(f, 0, np.inf)[0]
            elif z == 0:
                return 1
            else:
                return MLf(z, a)

        return np.vectorize(_MLf)(z, a)

    return A * MLf(-((t / tau) ** alpha), alpha)
